import numpy as np
import pylab
try:
    from enthought.mayavi import mlab
except  ImportError:
    mlab = None

from scipy.stats import chi2

from dirsdist.linalg \
    import convert_geo_to_angles, convert_geo_to_direction, generate_angles


def load_tracer_data(filename):
    """
    Loads data from tracer file.
    
    Parameters
    ----------
    - filename : str
    
    Returns
    -------
    - tracer_data : array
    - n_tracer_data : int
    """
    fd = open(filename, 'r')
    aux = fd.readline()
    n_tracer_data = int(fd.readline())
    print 'number of tracer data:', n_tracer_data

    tracer_data = np.zeros((n_tracer_data, 4), dtype=np.int)
    for ii in xrange(n_tracer_data):
        line = fd.readline()

        tracer_data[ii] = [int(aux) for aux in line.strip().split()]
    
    
    # tracer_data = np.loadtxt(fd, dtype=np.int)[:n_tracer_data]
    # assert(tracer_data.shape == (n_tracer_data, 4))

    print 'tracer data:'
    print tracer_data
    
    return n_tracer_data, tracer_data

def read_histograms(filename, vs, hs = None):
    """
    Reads histograms.
    
    Parameters
    ----------
    - filename : str
    - vs, hs : int
    
    Returns
    -------
    - data : array
    """
    if hs is None:
        hs = vs
    
    for ii in xrange(vs.shape[0]):
        name = filename % (vs[ii], hs[ii])
        print name
        data = np.loadtxt(name)
        yield data

def histogram_mean(fig, data, title_prefix=None):
    """
    Counts and draws histogram: distribution of lengths of blood vessels.
    
    Parameters
    ----------
    - fig : int
    - data : array
    - title_prefix : str
    """
    
    v_elev = data[:,0].max() + 1
    h_azim = data[:,1].max() + 1
    values = data[:,2]
    z = np.reshape(values, (h_azim, v_elev))
    
    elev, azim = generate_angles(v_elev, h_azim) 
    
    elev_mean = elev[:-1] + 0.5 * np.diff(elev)
    azim_mean = azim[:-1] + 0.5 * np.diff(azim)
    
    #print 'azim', azim
    #print 'azim_mean', azim_mean
    #print 'elev_mean', elev_mean
    #print 'z', z
    
    if (len(azim_mean) ==1) or (len(elev_mean)==1):
        if len(azim_mean) ==1:
            pylab.figure(fig)
            pylab.clf()
            pylab.plot(elev_mean, z)
            pylab.xlabel('elevation')
            pylab.ylabel('length')
            pylab.xticks( np.round( elev_mean ) )
            if title_prefix is None:
                pylab.title( 'Histogram v=%d, h=%d' % (v_elev, h_azim) )
            else:
                pylab.title( 'Histogram (%s) v=%d, h=%d' % (title_prefix, v_elev, h_azim) )
        else:
            pylab.figure(fig)
            pylab.clf()
            pylab.plot(azim_mean, z)
            pylab.xlabel('azimuth')
            pylab.ylabel('length')
            pylab.xticks( np.round( azim_mean ) )
            if title_prefix is None:
                pylab.title( 'Histogram v=%d, h=%d' % (v_elev, h_azim) )
            else:
                pylab.title( 'Histogram (%s) v=%d, h=%d' % (title_prefix, v_elev, h_azim) )
    else:
        pylab.figure(fig)
        pylab.clf()
        pylab.contourf( azim_mean, elev_mean, z, cmap = pylab.cm.hot_r )
        pylab.colorbar( format = '%d')
        pylab.xlabel( 'azimuth' )
        pylab.ylabel( 'elevation' )
        pylab.xticks( np.round( azim_mean ) )
        pylab.yticks( np.round( elev_mean ) )
        if title_prefix is None:
            pylab.title( 'Histogram v=%d, h=%d' % (v_elev, h_azim) )
        else:
            pylab.title( 'Histogram (%s) v=%d, h=%d' % (title_prefix, v_elev, h_azim) )

def histogram_mean_fraction(fig, data, fraction, title_prefix=None):
    """
    Counts and draws histogram: distribution of fractions of lengths of 
    blood vessels, i.e. :math:`l_i/\sum_{n=1}^N l_i`, where :math:`l_i` is the length of
    blood vessels in given direction, :math:`N` is total number of directions.
    
    Parameters
    ----------
    - fig : int
    - data : array
    - fraction : array
    - title_prefix : str
    """
    v_elev = data[:,0].max() + 1
    h_azim = data[:,1].max() + 1
    values = fraction
    z = np.reshape(values, (h_azim, v_elev))
    
    elev, azim = generate_angles(v_elev, h_azim) 
    
    elev_mean = elev[:-1] + 0.5 * np.diff(elev)
    azim_mean = azim[:-1] + 0.5 * np.diff(azim)
        
    pylab.figure(fig)
    pylab.clf()
    pylab.contourf( azim_mean, elev_mean, z, cmap = pylab.cm.hot_r )
    pylab.colorbar( format = '%.3f')
    pylab.xlabel( 'azimuth' )
    pylab.ylabel( 'elevations' )
    pylab.xticks( np.round( azim_mean ) )
    pylab.yticks( np.round( elev_mean ) )
    if title_prefix is None:
        pylab.title( 'Histogram of fractions v=%d, h=%d' % (v_elev, h_azim) )
    else:
        pylab.title( 'Histogram of fractions (%s) v=%d, h=%d' % (title_prefix, v_elev, h_azim) )

def count_chi_square(data, fraction):
    r"""
    Chi-square test of fractions distribution: 
    :math:`\chi^2 = \sum_{n=1}^N \frac{O_i-E_i}{E_i}`, where :math:`O_i` is observed 
    value, :math:`E_i` is expected value, :math:`N` is total number of values. In our case 
    :math:`E_i` is value of fraction when we assumed uniform distribution, 
    i.e. :math:`E_i = E_{i+1} =...= E_{N} = 1/v h`, where :math:`v` = number of elevation 
    divisions and :math:`h` = number of azimuth divisions.
    
    Parameters
    ----------
    - data : array
    - fraction : array
    
    Returns
    -------
    - chi_square : float
    """
    total_number = data.shape[0]
    value_expected = 1.0/total_number
    
    chi_square_elem = np.square(fraction - value_expected)
    
    sum_chi_square = np.sum(chi_square_elem)
    chi_square = total_number * sum_chi_square
    
    print 'chi_square', chi_square
    
    return chi_square
    

def evaluate_chi_square(chi_square):
    """
    Evaluates value of chi-square with respect to the probability value:
    :math:`p < 0.05` = anisotropic distribution, :math:`p > 0.05` = isotropic distribution.
    
    Parameters
    ----------
    - chi_square : float
    
    Returns
    -------
    - p_val : float
    """
    
    data_chi2 = np.array(range(0, 100000), float)/1000
    pdf_chi2 = chi2.pdf(data_chi2, 1) # probability density function df=1
    
    p_val_index = np.where(data_chi2 > chi_square)[0][0]
    print data_chi2[p_val_index], chi_square
    p_val = pdf_chi2[p_val_index]
    
    if p_val <= 0.001:
        msg = '*** (p<0.001) = the most significant, \alpha = ', p_val
        print msg
    elif (p_val > 0.001) & (p_val <= 0.01):
        msg = '** (p<0.01) = more significant, \alpha =', p_val
        print msg
    elif (p_val > 0.01) & (p_val <= 0.05):
        msg = '* (p<0.05) = significant, \alpha = ', p_val
        print msg
    else:
        if p_val >= 1: 
            p_val = 1
            print r'(p>0.05 (n.s.)) = distribution is isotropic, \alpha = ', p_val
        else:
            print r'(p>0.05 (n.s.)) = distribution is isotropic, \alpha = ', p_val
    
    return p_val

def evaluate_anisotropy(n_tracer_data, tracer_data, calibration):
    """
    Evaluates fractional and ellipsoidal anisotropy from the line system in file tracer. 
    The ellipsoidal anisotropy has the values from 1 (isotropy - lines express no 
    preferential direction) to infinity (total anisotropy - all lines run in the same direction). 
    The fractional anisotropy has the values from 0 (isotropy) to 1 (anisotropy). 
    
    Parameters
    ----------
    - n_tracer_data: int
    - tracer_data : array
    - calibration: array
    
    Returns
    -------
    - fractional_anisotropy : float
    - ellipsoidal_anisotropy : float
    """
    coors = np.empty((n_tracer_data, 3), dtype=np.float64)
    for ic in range(3):
        coors[:,ic] = tracer_data[:,ic] * calibration[ic]

    flags = np.diff(tracer_data[:,3])
    print 'flags:'
    print flags

    print 'coordinates:'
    print coors

    vectors = coors[1:] - coors[:-1]
    vectors = vectors[flags == 0]

    print 'vectors:'
    print vectors

    lengths = np.sqrt(np.sum(vectors**2, axis=1))

    print 'lengths:'
    print lengths

    directions = vectors / lengths[:,None]

    print 'directions:'
    print directions

    n_seg = directions.shape[0]

    matrix = np.zeros((3, 3), dtype=np.float64)

    for ii, v0 in enumerate(vectors):
        v1 = directions[ii]
        matrix += np.outer(v0, v1)

    print 'average matrix:'
    print matrix

    eigs, eig_vecs = np.linalg.eig(matrix)

    ii = np.argsort(eigs)
    eigs = eigs[ii]
    eig_vecs = eig_vecs[:,ii]

    np.set_printoptions(precision=3)

    print 'eigenvalues:', eigs
    print 'eigenvectors:'
    print eig_vecs

    seigs = np.sqrt(eigs)
    print 'principal axes lengths', seigs

    np.set_printoptions()

    ellipsoidal_anisotropy = 2.0 * seigs[2] / (seigs[0] + seigs[1])

    assert(ellipsoidal_anisotropy >= 1.0)

    print 'ellipsoidal anisotropy:', ellipsoidal_anisotropy

    avg_eig = eigs.sum() / 3.0

    val = np.sum((eigs - avg_eig)**2) / np.sum(eigs**2)
    val = np.sqrt(1.5 * val)
    fractional_anisotropy = val

    assert((fractional_anisotropy >= 0.0) and (fractional_anisotropy <= 1.0))

    print 'fractional anisotropy:', fractional_anisotropy
    
    evaluate_fractional_anisotropy(fractional_anisotropy)
    evaluate_ellipsoidal_anisotropy(ellipsoidal_anisotropy)
    
    return ellipsoidal_anisotropy, fractional_anisotropy


def evaluate_fractional_anisotropy( FA ):
    r"""
    Evaluates value of fractional anisotropy (FA): FA = 0 = isotropic system, 
    FA :math:`\approx` 1 = anisotropic system with very strong preferential directions.
        
    Parameters
    ----------
    - fractional_anisotropy
    
    """
    if FA == 0:
        msg = ' fractional isotropy = 0, the system is totally isotropic'
        print msg
    elif (FA> 0) & (FA <1):
        msg = ' the system is anisotropic with value of fractional isotropy = ' , FA
        print msg
    else:
        msg = ' the system is highly anisotropic with preferential directions, value of fractional isotropy = ' , FA
        print msg        

def evaluate_ellipsoidal_anisotropy( EA ):
    """
    Evaluates value of ellipsoidal aninisotropy (EA): ' EA = 1 = isotropic system, 
    EA = infinity = anisotropic system with very strong preferential directions.
        
    Parameters
    ----------
    - ellipsoidal_anisotropy
    
    """
    
    if EA == 1:
        msg = ' (ellipsoidal anisotropy = 1) = the system is totally isotropic'
        print msg
    else:
        msg = ' the system is anisotropic with value of ellipsoidal isotropy = ' , EA
        print msg        


def sort_data(data):
    """
    Sorts data in ascending manner.
    
    Parameters
    ----------
    - data : array
    
    Returns
    -------
    - data_sorted : array
    """
    i_sort = data[:,2].argsort()
    data_sorted = data[i_sort,:]
    
    return data_sorted
    

def compute_fraction(data):
    """
    Computes fractions of lengths of blood vessels in all directions:
    :math:`l_i/\sum_{n=1}^N l_i`, where :math:`l_i` is the length of blood vessels in 
    given direction, :math:`N` is total number of directions.    
    
    Parameters
    ----------
    - data : array
    
    Returns
    -------
    - fraction : array
    """
    sum_lengths = np.sum(data[:,2])
    fraction = data[:,2] / sum_lengths
    
    return fraction
    

def angle_max (data_sorted, n_max):
    """
    Finds `n_max` directions of blood vessels with the highest length densities.
    
    Parameters
    ----------
    - data_sorted : array
    - n_max : int
    
    Returns
    -------
    - (h_azim, v_elev) : array
    -  vector : array
    -  vector_mean : array
    """
    v_elev = data_sorted[:,0].max()+1
    h_azim = data_sorted[:,1].max()+1
    
    
    value_v_max=data_sorted[-n_max:,0]; # value of v for maxima (number = n_max)
    value_h_max=data_sorted[-n_max:,1]  # value of h for maxima (number = n_max)
    
    elevs, azims = convert_geo_to_angles([value_v_max, value_v_max+1],
                                                            [value_h_max, value_h_max+1], 
                                                            v_elev, h_azim)

    # elevs[0] = initial elevation of interval with cardinal direction
    # elevs[1] = final elevation of interval with cardinal direction
    # analogously for azims
    
    elev_mean = np.sum(elevs, axis=0)/ 2.0 # mean value of interval of elevation
    azim_mean = np.sum(azims, axis=0)/ 2.0 # mean value of interval of elevation
    
    vector = [azims[0], azims[1], elevs[0], elevs[1]]  # final vector with main directions for n_max
    vector_mean = [azim_mean, elev_mean] # final vector with middle of intervals with main directions for n_max
    print 'elevs:'
    print elevs
    print 'azims:'
    print azims
    print 'vector:'
    print vector
    print 'vector_mean:'
    print vector_mean
    
    return (h_azim, v_elev), vector, vector_mean

def find_n_max(fraction, threshold):
    """
    Finds the number of directions necesarry to covered threshold value,
    i.e. given fraction of length densities :math:`[0,1]`.
    
    Parameters
    ----------
    - fraction : array
    - threshold : float
    
    Returns
    -------
    -  n_max : int
    """
    ii_0 = ii = len(fraction) - 1
    sum_fraction = 0.0
    
    while ii > 0:
        #print ii, sum_fraction
        sum_fraction += fraction[ii]
        if sum_fraction >= threshold:
            break
        ii += -1
    n_max = ii_0 - ii + 1
    
    print 'n_max', n_max
    return n_max


def evaluate_isotropy(data):
    """
    Evaluates isotropy:
    
    Isotropic means "independent of direction". Isotropic radiation has 
    the same intensity regardless of the direction of measurement, and 
    an isotropic field exerts the same action regardless of how the test 
    particle is oriented. This evaluation is based on the chi-square test; 
    null hypothesis :math:`H_0` the observed fractions have uniform 
    distribution (through all directions).
    
    Parameters
    ----------
    - data : array
    
    Returns
    -------
    - chi_square : float
    - p_val : float
    - mean_fraction : float
    - std_fraction : float
    """
    fraction = compute_fraction(data)
    
    mean_fraction = np.mean(fraction)
    std_fraction = np.std(fraction)
    
    chi_square = count_chi_square(data, fraction)
    
    p_val = evaluate_chi_square(chi_square)
        
    print 'fraction ditribution: mean \pm SD', mean_fraction, '\pm', std_fraction
    
    return chi_square, p_val, mean_fraction, std_fraction
        
def rose (fig, data):
    """
    Draws original rose of directions based on dividing given by :math:`v` and :math:`h`
    (elevations and azimuths).
    
    Parameters
    ----------
    - fig : int
    - data : array
    """
    v_elev = data[:,0].max() + 1
    h_azim = data[:,1].max() + 1
    # initial coordinates of individual directions of rose of direction =
    # zeros, the vectors has origin in zero
    [data_rows, data_columns]=np.shape( data ) 
    x10=np.zeros( data_rows, dtype=float) # vector with initial values = zeros
    x20=np.zeros( data_rows, dtype=float) # vector with initial values = zeros
    x30=np.zeros( data_rows, dtype=float) # vector with initial values = zeros
       
    # conversion from v,h to coordinates of vector in half cell (spherical tetragon)
    x1, x2, x3 = convert_geo_to_direction(data[:,0]+0.5, data[:,1]+0.5, v_elev, h_azim)
    
    mlab.figure(fig)
    mlab.clf()
    mlab.quiver3d(x10,x20,x30,x1,x2,x3)    #upper hemisphere
    mlab.quiver3d(x10,x20,x30,-x1,-x2,-x3)    #bottom hemisphere
    #mlab.title('Rose of Directions')
    #mlab.xlabel('x')
    #mlab.ylabel('y')
    #mlab.zlabel('axis perpendicular to sections')
    

def rose_scaled (fig, data):
    """
    Draws scaled rose of directions based on dividing given by :math:`v` and :math:`h`
    (elevations and azimuths).
    
    Parameters
    ----------
    - fig : int
    - data : array
    """
    v_elev = data[:,0].max() + 1
    h_azim = data[:,1].max() + 1
    # initial coordinates of individual directions of rose of direction =
    # zeros, the vectors has origin in zero
    [data_rows, data_columns]=np.shape( data ) 
    x10=np.zeros( data_rows, dtype=float) # vector with initial values = zeros
    x20=np.zeros( data_rows, dtype=float) # vector with initial values = zeros
    x30=np.zeros( data_rows, dtype=float) # vector with initial values = zeros
   
    # conversion from v,h to coordinates of vector in half cell (spherical tetragon)
    x1, x2, x3 = convert_geo_to_direction(data[:,0]+0.5, data[:,1]+0.5, v_elev, h_azim)
    
    # coordinates of vector in half cell (spherical tetragon) multiplied by length of vector 
    x1_length=data[:,2]*x1
    x2_length=data[:,2]*x2
    x3_length=data[:,2]*x3
    
    mlab.figure(fig)
    mlab.clf()
    h = mlab.quiver3d(x10,x20,x30,x1_length,x2_length,x3_length)   # upper hemisphere
    #import pdb; pdb.set_trace()
    h.glyph.glyph.clamping = 0
    h = mlab.quiver3d(x10,x20,x30,-x1_length,-x2_length,-x3_length) # bottom hemisphere
    h.glyph.glyph.clamping = 0
       
    mlab.title('Scaled Rose of Directions')
    mlab.xlabel('x')
    mlab.ylabel('y')
    mlab.zlabel('axis perpendicular to sections')
    xmax = np.abs( x1_length ).max()
    ymax = np.abs( x2_length ).max()
    zmax = np.abs( x3_length ).max()
    ranges = [-xmax, xmax, -ymax, ymax, -zmax, zmax]
    ax = mlab.axes(ranges = ranges, extent=ranges)
    ax.axes.number_of_labels = 4
        
def lambert_azimuthal_equal_area( az, el ):
    """
    Counts lambert axes.
    
    Parameters
    ----------
    - az : int
    - el : int
    
    Returns
    -------
    - x_lambert : float
    - y_lambert : float
    """ 
    # elevation shifted by pi - (pi/4 + el)!
    r = np.sqrt( 2 * (1 + np.cos( np.pi - (np.pi / 4 + el * np.pi / 180.) )) )
    phi = az * np.pi / 180

    x_lambert = r * np.cos( phi )
    y_lambert = r * np.sin( phi )
    
    return x_lambert, y_lambert

def plot_lambert_axes( fig, x, txt ):
    """
    Plots lambert axes.
    
    Parameters
    ----------
    - fig : int
    - x : int
    - txt : int
    """ 
    # elevation shifted by pi - (pi/4 + el)!
    mlab.figure( fig )
           
    elevations = (0, 30, 60, 90)
    n_az = 21
    az = np.linspace( 0, 360, n_az )
      
    phi = az * np.pi / 180
    for el in elevations:
        r = np.sqrt( 2 * (1 + np.cos( np.pi - (np.pi/4 + el * np.pi / 180.) )) )
        y = r * np.cos( phi )
        z = r * np.sin( phi )
        h = mlab.plot3d( np.repeat( x, n_az ), y, z,  color = (0.0, 0.0, 0.0), tube_radius = 0.01)
         
        if txt:
            yt = (r + 0.1) * np.cos( np.pi/8 )
            zt = (r + 0.1) * np.sin( np.pi/8 )
            mlab.text(x, yt, '%02d' % (el), z=zt, width=0.03, color=(0,0,0)) 
     
    azimuths = (0, 45, 90, 135, 180, 225, 270, 315)
    for az in azimuths:
        phi = az * np.pi / 180
        yt = 2 * np.cos( phi )
        zt = 2 * np.sin( phi )
        #h = mlab.plot3d( [x, x], [0, yt], [0, zt], color = (0.0, 1.0, 0.0))
        h = mlab.plot3d( [x, x], [0, yt], [0, zt], color = (0.75,0.75,0.75), tube_radius = 0.01)
        
        if txt:
            mlab.text(x, yt, '%03d' % (az), z=zt, width=0.05, color=(0,0,0)) 
    
def cardinal_orientation(fig, data_sorted, ii, n_max, plot_axes = False):
    """
    Plots `n_max` cardinal directions.
    
    Parameters
    ----------
    - fig : int
    - data_sorted : array
    - ii : int
    - n_max : int
    - plot_axes : bool
    
    Returns
    -------
    - v1 : array
    - h1 : array
    """ 
    ((h_azim, v_elev), vector, vector_mean) = angle_max( data_sorted, n_max )
     
    n_hist = np.size(vector_mean, 0)/2
            
    x0=np.array( [ii] * n_max, dtype = float)
    h0=np.zeros(n_max, dtype=float)
    dx = v0 = h0
    
    mlab.figure(fig, bgcolor = (1.0, 1.0, 1.0), size=(800,800) )
                 
    if plot_axes:
        plot_lambert_axes( fig, ii, 1 )
        
    h1 = vector_mean[0]
    v1 = vector_mean[1]
         
    (x_lambert, y_lambert) = lambert_azimuthal_equal_area( h1, v1 )
    #import pdb; pdb.set_trace()
    h = mlab.quiver3d( x0, v0, h0, dx, x_lambert, y_lambert, color=(0.0,0.0,0.0))
    #h=mlab.quiver3d( x0, v0, h0, dx, x_lambert, y_lambert )
    h.glyph.glyph.clamping = 0
     
    return v1, h1 
