import numpy as np

def make_axis_rotation_matrix(direction, angle):
    r"""
    Creates a rotation matrix corresponding to the rotation around a general
    axis by a specified angle.
    
    :math:`R = \vec{d}\vec{d}^T + \cos(a) (I - \vec{d}\vec{d}^T) + \sin(a) skew(\vec{d})`
    
    Parameters
    -----------------
    - direction : array
    - angle : float
    
    Returns
    -----------
    - mtx : array
    """
    d = np.array(direction, dtype=np.float64)
    d /= np.linalg.norm(d)
    
    eye = np.eye(3, dtype=np.float64)
    ddt = np.outer(d, d)
    skew = np.array([[    0,  d[2],  -d[1]],
                     [-d[2],     0,  d[0]],
                     [d[1], -d[0],    0]], dtype=np.float64)

    mtx = ddt + np.cos(angle) * (eye - ddt) + np.sin(angle) * skew
    return mtx

def convert_angles_to_direction(elev, azim):
    r"""
    Converts elevation and azimuth angles to direction vectors.
    
    Parameters
    -----------------
    - elev : float
    - azim : float
    
    Returns
    -----------
    - direction : array
    """
    elev = np.pi * elev / 180.0
    azim = np.pi * azim / 180.0
    
    x1 = np.cos(elev) * np.cos(azim)
    x2 = np.cos(elev) * np.sin(azim)
    x3 = np.sin(elev)
    
    return x1, x2, x3

def convert_geo_to_direction(v, h, v_elev, h_azim):
    """
    Converts from v, h to coordinates of vector in half cell (spherical tetragon). 
    
    Parameters
    -----------------
    - v : int
    - h : int
    - v_elev : int
    - h_azim : int
    
    Returns
    -----------
    - direction : array
    """
    v = np.asarray(v, dtype=np.float64) / v_elev
    h = np.asarray(h, dtype=np.float64) / h_azim
 
    x3 = v
    x1 = np.sqrt(1.0-np.square(v)) * np.cos(2.0*np.pi*h)
    x2 = np.sqrt(1.0-np.square(v)) * np.sin(2.0*np.pi*h)
    
    return x1, x2, x3

def convert_geo_to_angles(v, h, v_elev, h_azim):
    """
    Converts from v, h to elevation and azimuth angles. 
    
    Parameters
    -----------------
    - v : int
    - h : int
    - v_elev : int
    - h_azim : int
    
    Returns
    -----------
    - elev : float
    - azim : float
    """
    
    v = np.asarray(v, dtype=np.float64)
    h = np.asarray(h, dtype=np.float64)
    
    elev = 180.0 * np.arcsin(v/v_elev) / np.pi
    azim = 360.0 * h / h_azim
    
    return elev, azim

def generate_angles(v_elev, h_azim):
    """
    Generates uniformly spaced elevation and azimuth angles 
    given the number of parallels and meridians.
    
    Parameters
    -----------------
    - v_elev : int
    - h_azim : int
    
    Returns
    -----------
    - elev : float
    - azim : float
    """
    
    ii = np.arange(v_elev + 1 , dtype = np.float64 )
    jj = np.arange(h_azim + 1, dtype = np.float64)
    
    return convert_geo_to_angles(ii, jj, v_elev, h_azim)
